import { Component, Inject } from '@angular/core';
import { HttpClient } from '@angular/common/http';

@Component({
  selector: 'app-calendar',
  templateUrl: './calendar.component.html'
})
export class CalendarComponent {
  public forecasts: CalendarForecast[];
  public webSocket;
  public calendars;

  getCalendar(id) {
    console.log("getCalendar: ", id);
    this.webSocket.send('{Op: "calget", ChannelId: ' + id + '}');
  }

  handleCalendarList(cList) {
    console.log("Setting calendars: ", cList);
    this.calendars = cList;
  };

  constructor(http: HttpClient, @Inject('BASE_URL') baseUrl: string) {
    var self = this;
    console.log("ws connecting to ws" + baseUrl.substring(4, baseUrl.length) + "ws/calendar");
    this.webSocket = new WebSocket("ws" + baseUrl.substring(4, baseUrl.length) + "ws/calendar");
    

    this.webSocket.onopen = function () {
      // TODO: Stop hardcoding gid, get this from url?
      self.webSocket.send('{Op:"connect", GId: 98293701175955456, Module:"calendar"}');
      setInterval(() => {
        self.webSocket.send('{Op:"ping"}');
      },120000);
    };

    this.webSocket.onmessage = function (evt) {
      console.log("evt: ", evt);
      var msg = JSON.parse(evt.data);
      console.log(msg);
      switch (msg.Op) {
        case "CalendarList":
          self.handleCalendarList(msg.Calendars);
          break;
        default:
          console.log("Unknown opcode: ", msg);
      }
    };

    this.webSocket.onclose = function () {

      // websocket is closed.
      alert("Connection is closed...");
    };
  }
}

interface CalendarForecast {
  dateFormatted: string;
  temperatureC: number;
  temperatureF: number;
  summary: string;
}
