﻿using DiscoBot.calendar;
using Discord;
using Discord.WebSocket;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Dynamic;
using System.IO;
using System.Linq;
using System.Net.WebSockets;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace DiscoBot
{
    public class DisBot
    {
        private DiscordSocketClient discordClient;
        private WSController wsC;
        private List<Type> moduleTypes = new List<Type>();
        private Dictionary<ulong, List<IModule>> guildModules = new Dictionary<ulong, List<IModule>>();
        private Dictionary<ulong, Dictionary<string, Func<SocketMessage, string[], Task>>> guildcommands = new Dictionary<ulong, Dictionary<string, Func<SocketMessage, string[], Task>>>();
        private CalendarContext calendarContext;
        public DisBot(WSController wsC)
        {
            this.wsC = wsC;
        }

        public async Task HandleWebsocket(HttpContext context, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Handling Websocket to "+ context.Request.Path);
            WebSocket webSocket = await context.WebSockets.AcceptWebSocketAsync();
            string msg = await DisBot.ReadOneStringFromWebSocket(webSocket);
            var login = JsonConvert.DeserializeObject<Login>(msg);
            var modules = guildModules[login.Gid];
            foreach (var m in modules)
            {
                if (m.Name == login.Module)
                {
#pragma warning disable CS4014 // Da auf diesen Aufruf nicht gewartet wird, wird die Ausführung der aktuellen Methode vor Abschluss des Aufrufs fortgesetzt.
                    m.OnNewWebSocketAsync(webSocket, tcs);
#pragma warning restore CS4014 // Da auf diesen Aufruf nicht gewartet wird, wird die Ausführung der aktuellen Methode vor Abschluss des Aufrufs fortgesetzt.
                }
            }
        }

        public static async Task<string> ReadOneStringFromWebSocket(WebSocket ws)
        {
            ArraySegment<Byte> buffer = new ArraySegment<byte>(new Byte[8192]);
            WebSocketReceiveResult result = await ws.ReceiveAsync(buffer, CancellationToken.None);
            var buf = buffer.ToArray();
            string line = System.Text.Encoding.UTF8.GetString(buf, 0, result.Count);
            Console.WriteLine("READ: " + line + "...:::...");
            return line;
        }

        public static async void SendStringViaWebSocket(WebSocket ws, string str)
        {
            var msgBytes = Encoding.UTF8.GetBytes(str);
            Console.WriteLine("Sending: "+str);
            await ws.SendAsync(new ArraySegment<byte>(msgBytes, 0, msgBytes.Length), WebSocketMessageType.Text, true, CancellationToken.None);
        }

        public async void Initialize()
        {
            Console.WriteLine("In init!");
            discordClient = new DiscordSocketClient();
            
            // DiscoBot
            await discordClient.LoginAsync(TokenType.Bot, "NTUxNDcxODcxNDcyNjk3MzQ1.D3DVtw.Weh-a3l2XsBGuD5N1-rLKfnZ8wI");
            // DiscoBot_beta
            //await discordClient.LoginAsync(TokenType.Bot, "NjM0ODUzOTUzMTY2OTAxMjgw.XaokNQ.Pja5GGV-Hz17TbO7y9WZ6YuPHZE");
            await discordClient.StartAsync();

            discordClient.Log += Log;
            discordClient.JoinedGuild += JoinedGuild;
            discordClient.GuildAvailable += GuildAvailable;
            discordClient.Connected += Connected;
            discordClient.Ready += Ready;
            discordClient.LeftGuild += LeftGuild;
            discordClient.MessageReceived += MessageReceived;

            Console.WriteLine("Looking for modules...");
            Assembly[] assemblies = AppDomain.CurrentDomain.GetAssemblies();
            foreach(Assembly a in assemblies)
            {
                moduleTypes.AddRange(GetTypesWithInterface(a));
            }
            Console.WriteLine("Found " + moduleTypes.Count + " module types.");
        }

        private Task Log(LogMessage msg)
        {
            Console.WriteLine(msg.ToString());
            return Task.CompletedTask;
        }

        private Task Connected()
        {
            Console.WriteLine("CONNECTED WOHOOOO");
            return Task.CompletedTask;
        }
        private Task Ready()
        {
            Console.WriteLine("Ready, connected to the following guilds: ");
            foreach(SocketGuild guild in discordClient.Guilds)
            {
                Console.WriteLine("- "+guild.Name+" | "+guild.Id);
            }
            return Task.CompletedTask;
        }

        private Task JoinedGuild(SocketGuild guild)
        {
            Console.WriteLine("Joined Guild "+guild.Name);
            return Task.CompletedTask;
        }
        
        private void LoadModulesForGuild(SocketGuild guild)
        {
            Dictionary<string, Func<SocketMessage, string[], Task>> commands = new Dictionary<string, Func<SocketMessage, string[], Task>>();
            List<IModule> iModules = new List<IModule>();
            foreach (Type t in moduleTypes)
            {
                if (t.IsInterface)
                    continue;
                try
                {
                    IModule module = (IModule)Activator.CreateInstance(t, new object[] { guild });
                    module.Initialize();
                    iModules.Add(module);
                    var c = module.Commands;
                    List<string> keyList = new List<string>(c.Keys);
                    foreach (string k in keyList)
                    {
                        Console.WriteLine("Found Command: " + k);
                    }
                    foreach (var entry in c)
                    {
                        commands.Add(entry.Key, entry.Value);
                    }
                } catch(MissingMethodException)
                {
                    Console.WriteLine("ERROR: Could not instantiate " + t.FullName + ". Could not find SocketGuild constructor.");
                }
            }
            guildcommands.Add(guild.Id, commands);
            guildModules.Add(guild.Id, iModules);
        }

        private Task GuildAvailable(SocketGuild guild)
        {
            Console.WriteLine("Guild available " + guild.Name + " | " + guild.Id);
            if (guildcommands.ContainsKey(guild.Id))
            {
                Console.WriteLine("Guild " + guild.Id + " is already loaded.");
                return Task.CompletedTask;
            }
            LoadModulesForGuild(guild);
            return Task.CompletedTask;
        }

        private Task LeftGuild(SocketGuild guild)
        {
            Console.WriteLine("Left Guild " + guild.Name);
            return Task.CompletedTask;
        }

        private Task MessageReceived(SocketMessage message)
        {
            // Don't parse our own messages
            if (message.Author.Id == discordClient.CurrentUser.Id)
                return Task.CompletedTask;
            
            if(message.Content.StartsWith("!"))
            {
                if(message.Channel is IGuildChannel)
                {
                    var gchan = message.Channel as IGuildChannel;
                    Console.WriteLine("|"+gchan.Name+"| <" + message.Author + "> " + message.Content);
                    Dictionary<string, Func<SocketMessage, string[], Task>> commands = guildcommands[gchan.GuildId];
                    string cnt = message.Content.Substring(1);
                    string[] splits = cnt.Split(" ");
                    string command = splits[0];

                    return Task.Run(() => {
                        if(commands.ContainsKey(command))
                            commands[command](message, splits);
                        else
                            Console.WriteLine("Command not found!");
                    });
                } else
                {
                    Console.WriteLine("Out of guild command! "+message.Content);
                    message.Channel.SendMessageAsync("Commands in DM are not supported. :(");
                }
            }
            return Task.CompletedTask;
        }

        private IEnumerable<Type> GetTypesWithInterface(Assembly asm)
        {
            var it = typeof(IModule);
            return asm.GetLoadableTypes().Where(it.IsAssignableFrom).ToList();
        }
    }
    public static class TypeLoaderExtensions
    {
        public static IEnumerable<Type> GetLoadableTypes(this Assembly assembly)
        {
            if (assembly == null) throw new ArgumentNullException("assembly");
            try
            {
                return assembly.GetTypes();
            }
            catch (ReflectionTypeLoadException e)
            {
                return e.Types.Where(t => t != null);
            }
        }
    }
}