﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Net.WebSockets;
using System.ServiceModel.Syndication;
using System.Threading.Tasks;
using System.Timers;
using System.Xml;
using Discord;
using Discord.WebSocket;
using Microsoft.EntityFrameworkCore;

namespace DiscoBot.bdaygreet
{
    public class BDayGreet : IModule
    {
        string IModule.Name { get => "BDayGreet"; set => throw new NotImplementedException(); }
        private SocketGuild guild;
        private BDayGreetContext bdayGreetContext;
        private Timer updateTimer = new Timer();
        private List<WebSocket> webSockets = new List<WebSocket>();

        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        public BDayGreet(SocketGuild guild)
        {
            this.guild = guild;
            bdayGreetContext = new BDayGreetContext(guild.Id);
            bdayGreetContext.Database.EnsureCreated();
            InitializeCheck();

            Commands.Add("bdayadd", HandleBDayAddCommand);
            Commands.Add("bdaydel", HandleBDayDelCommand);
            Commands.Add("bdaylist", HandleBDayListCommand);
            Commands.Add("bdaydebug", HandleBDayDebugCommand);
        }

        private void SetTimer()
        {
            updateTimer.Dispose();
            var now = DateTime.Now;
            //DateTimeOffset next = now.AddHours(1).AddMinutes(-now.Minute).AddSeconds(-now.Second);
            DateTimeOffset next = now.AddMinutes(1).AddSeconds(-now.Second);
            var span = next - now;
            Console.WriteLine("Next bday timer is: " + next + " in " + span.TotalMilliseconds + "ms.");
            updateTimer = new Timer(span.TotalMilliseconds);
            updateTimer.AutoReset = false;
            updateTimer.Elapsed += async (sender, e) =>
            {
                try
                {
                    await Task.Run(() => HandleBDayCheck());
                } catch(Exception)
                {
                    Console.WriteLine("bday: EXCEPTION IN BDAY CHECK");
                }
            };
            updateTimer.Start();
        }

        private Task InitializeCheck()
        {
            Console.WriteLine("Starting check timer!");
            SetTimer();
            HandleBDayCheck();
            return Task.CompletedTask;
        }

        /*
        private Task DeinitializeFeed(BDayFeed feed)
        {
            Timer t = timers[feed.Name];
            t.Stop();
            timers.Remove(feed.Name);
            return Task.CompletedTask;
        }
        */
        private async Task HandleBDayCheck()
        {
            Console.WriteLine("Checktime!");
            var bdays = bdayGreetContext.BDays;
            foreach(var bday in bdays)
            {
                SocketTextChannel c = guild.Channels.Where(g => g.Id == bday.Channel).Single() as SocketTextChannel;
                //await c.SendMessageAsync("Checking " + bday.DiscordTag + "'s birthday. NextGreet: "+bday.NextGreet);
                var now = DateTimeOffset.Now;
                if(bday.NextGreet < now)
                {
                    var call = "";
                    try
                    {
                        call = guild.Users.Where(gu => (gu.Username + "#" + gu.DiscriminatorValue) == bday.DiscordTag).Single().Mention;
                    }
                    catch (InvalidOperationException e)
                    {
                        // Too spammy, do something about it? Don't rely on u.User
                        //c.SendMessageAsync("Could not find User "+ u.DiscordTag+" in Discord, cannot mention.");
                        call = bday.DiscordTag;
                    }
                    if (bday.NextGreet.Day == now.Day && bday.NextGreet.Month == now.Month)
                    {
                        await c.SendMessageAsync("Today is "+call+"'s birthday! Happy Birthday " + call+"!");
                    } else
                    {
                        await c.SendMessageAsync("Happy belated Birthday " + call+"! I guess I was offline when it happened. :(");
                    }
                    bday.NextGreet = bday.NextGreet.AddYears(1);
                    bdayGreetContext.SaveChanges();
                }
            }
            SetTimer();
        }

        private async Task HandleBDayAddCommand(SocketMessage msg, string[] parameters)
        {
            var gauthor = guild.Users.Where(u => u.Id == msg.Author.Id).Single();
            if(!gauthor.GuildPermissions.Administrator)
            {
                await msg.Channel.SendMessageAsync("Sorry only administrators can add birthdays :(");
                return;
            }
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            string date = parameters[2];
            SocketGuildUser user;
            try {
            var id = MentionUtils.ParseUser(name);
            user = guild.Users.Where(u => u.Id == id).Single();
            } catch(Exception)
            {
                await msg.Channel.SendMessageAsync("I can't find user " + name + " so I can't add them.");
                return;
            }
            DateTimeOffset d;
            try {
                d = DateTimeOffset.ParseExact(date,new string[] { "dd.MM.", "d.MM.", "dd.M", "d.M." }, null, DateTimeStyles.AssumeUniversal);
            } catch(FormatException) {
                await msg.Channel.SendMessageAsync("Sorry I can't parse your date. Please try again!");
                return;
            } 
            
            var bday = new BDayUser();
            bday.DiscordTag = user.Username + "#" + user.DiscriminatorValue;
            bday.Channel = msg.Channel.Id;
            bday.BirthDate = d;
            if(bday.BirthDate > DateTimeOffset.Now)
            {
                bday.NextGreet = bday.BirthDate;
            } else
            {
                if(bday.BirthDate.Day == DateTimeOffset.Now.Day && bday.BirthDate.Month == DateTimeOffset.Now.Month)
                {
                    bday.NextGreet = bday.BirthDate;
                } else
                {
                    bday.NextGreet = bday.BirthDate.AddYears(1);
                }
            }

            try
            {
                bdayGreetContext.Add(bday);
                bdayGreetContext.SaveChanges();
                await msg.Channel.SendMessageAsync("I remembered that " + name + "'s next birthday is on " + bday.NextGreet.ToString("dd.MM.yyyy"));
            }
            catch (InvalidOperationException)
            {
                await msg.Channel.SendMessageAsync("I can't add " + name + ". Looks like they are already in the database?");
                return;
            }
        }

        private async Task HandleBDayListCommand(SocketMessage msg, string[] parameters)
        {
            var gauthor = guild.Users.Where(u => u.Id == msg.Author.Id).Single();
            if (!gauthor.GuildPermissions.Administrator)
            {
                await msg.Channel.SendMessageAsync("Sorry only administrators can list birthdays :(");
                return;
            }
            var bdays = bdayGreetContext.BDays.Where(b => b.Channel == msg.Channel.Id).ToList();
            List<string> m = new List<string>();
            m.Add("I am tracking birthdays for: ");
            int cnt = 0;
            string str = "";
            foreach(var bday in bdays)
            {
                str += bday.DiscordTag + " ";
                cnt++;
                if(cnt > 4)
                {
                    m.Add(str);
                    str = "";
                }
            }
            m.Add(str);
            await msg.Channel.SendMessageAsync(string.Join("\n",m));
        }

        private async Task HandleBDayDelCommand(SocketMessage msg, string[] parameters)
        {
            var gauthor = guild.Users.Where(u => u.Id == msg.Author.Id).Single();
            if (!gauthor.GuildPermissions.Administrator)
            {
                await msg.Channel.SendMessageAsync("Sorry only administrators can delete birthdays :(");
                return;
            }
            var gchan = msg.Channel as IGuildChannel;
            string tag = ((IModule)this).ConcatParameters(parameters[1..]);
            try
            {
                var b = bdayGreetContext.BDays.Where(b => b.DiscordTag == tag).Single();
                bdayGreetContext.Remove(b);
                bdayGreetContext.SaveChanges();
                await msg.Channel.SendMessageAsync("Removed "+ b.DiscordTag+"'s birthday from the database.");
            } catch(InvalidOperationException)
            {
                await msg.Channel.SendMessageAsync("Could not find " + tag+"'s birthday.");
            }
        }

        private async Task HandleBDayDebugCommand(SocketMessage msg, string[] parameters)
        {
            await msg.Channel.SendMessageAsync("DEBUG!");
            string name = parameters[1];
            var id = MentionUtils.ParseUser(name);
            var user = guild.Users.Where(u => u.Id == id).Single();
            await msg.Channel.SendMessageAsync("name: " + name);
            await msg.Channel.SendMessageAsync("Found user with id" + user.Id+" and mention "+user.Mention+" Username: "+user.Username+" Discriminator: "+user.Discriminator+" DiscriminiatorValue: "+user.DiscriminatorValue);
        }

        public void Initialize()
        {
            Console.WriteLine("Initializing bdaygreet...");
        }

        public async Task OnNewWebSocketAsync(WebSocket ws, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Calendar " + guild.Id + " has a new websocket.");
            webSockets.Add(ws);
        }
    }
}
