﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.WebSockets;
using System.Threading.Tasks;
using Discord;
using Discord.WebSocket;
using Microsoft.EntityFrameworkCore;
using Newtonsoft.Json;

namespace DiscoBot.calendar
{
    public class Calendar : IModule
    {
        string IModule.Name { get => "calendar"; set => throw new NotImplementedException(); }
        private SocketGuild guild;
        private List<WebSocket> webSockets = new List<WebSocket>();
        public enum Attendance { Attending, Maybe }
        private CalendarContext calendarContext;
        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        public Calendar(SocketGuild guild)
        {
            this.guild = guild;
            calendarContext = new CalendarContext(guild.Id);
            calendarContext.Database.EnsureCreated();                
            
            Commands.Add("caltest", HandleTestCommand);
            Commands.Add("caladd", HandleAddEventCommand);
        }

        public void Initialize()
        {
            Console.WriteLine("Initializing calendar...");
        }

        public async Task OnNewWebSocketAsync(WebSocket ws, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Calendar " + guild.Id + " has a new websocket.");
            webSockets.Add(ws);
            // Send list of calendars(channels)
            //var calendar = calendarContext.ChannelCalendar.Include(cal => cal.Items).ToList();
            var cList = new CalendarList();
            //
            // TODO: Make all ulong -> String because js can't handle ulong
            //
            cList.Calendars = calendarContext.ChannelCalendar.ToList();
            string str = JsonConvert.SerializeObject(cList);
            Console.WriteLine(str);
            DisBot.SendStringViaWebSocket(ws, str);
#pragma warning disable CS4014 // Da auf diesen Aufruf nicht gewartet wird, wird die Ausführung der aktuellen Methode vor Abschluss des Aufrufs fortgesetzt.
            Task.Run(async () =>
              {
                  Console.WriteLine("Starting websocket handling...");
                  while (ws.State != WebSocketState.Closed)
                  {
                      string msg = await DisBot.ReadOneStringFromWebSocket(ws);
                      var op = JsonConvert.DeserializeObject<Opcode>(msg);
                      switch (op.Op)
                      {
                          case "calget":
                              var calGet = JsonConvert.DeserializeObject<CalGet>(msg);
                              Console.WriteLine("ChannelID:" + calGet.ChannelId);
                              var calendarWS = new CalendarWS();
                              var temp = calendarContext.ChannelCalendar.FirstOrDefault();
                              calendarWS.Op = calGet.Op;
                              DisBot.SendStringViaWebSocket(ws, JsonConvert.SerializeObject(calendarWS));
                              break;
                          default:
                              Console.WriteLine("Unknown opcode: " + op.Op);
                              break;
                      }
                  }
                  Console.WriteLine("Websocket handling stopped...");
                  tcs.TrySetResult(new object());
              });
#pragma warning restore CS4014 // Da auf diesen Aufruf nicht gewartet wird, wird die Ausführung der aktuellen Methode vor Abschluss des Aufrufs fortgesetzt.
        }

        private Task HandleTestCommand(SocketMessage msg, string[] parameters)
        {
            Console.WriteLine("Calendar: Handling test command!");
            string para = string.Join(",", parameters);
            msg.Channel.SendMessageAsync("Test succeeded. Params: " + para);
            return Task.CompletedTask;
        }

        private Task HandleAddEventCommand(SocketMessage msg, string[] parameters)
        {
            if(parameters.Length != 4)
            {
                msg.Channel.SendMessageAsync("Usage: caladd Name Date Comment");
                return Task.CompletedTask;
            }
            try
            {
                var gchan = msg.Channel as IGuildChannel;
                var calendar = calendarContext.ChannelCalendar.FirstOrDefault(cal => cal.ChannelId == gchan.Id);
                if(calendar == null)
                {
                    calendar = new ChannelCalendar();
                    calendar.ChannelId = gchan.Id;
                    calendar.Moderated = false;
                    calendarContext.Add(calendar);
                }
                string name = parameters[1];
                string date = parameters[2];
                string comment = parameters[3];
                CalendarItem calItem = new CalendarItem();
                calItem.Channel = gchan.Id;
                calItem.Name = name;
            
                calItem.Date = DateTimeOffset.Parse(date);
                calItem.Comment = comment;
                calItem.Attendance = new List<CalendarItemAttendance>();
                var attendance = new CalendarItemAttendance();
                attendance.Attendee = msg.Author.Id;
                attendance.Attending = Attendance.Attending;
                calItem.Attendance.Add(attendance);
                calendar.Items.Add(calItem);
                calendarContext.SaveChanges();
                msg.Channel.SendMessageAsync("Event "+ calItem.Name + " has been added.");
            }
            catch(FormatException)
            {
                msg.Channel.SendMessageAsync("Your date input was invalid. Try again with a valid date.");
            }



            return Task.CompletedTask;
        }

        public void OnNewWebsocket(WebSocket ws)
        {
            throw new NotImplementedException();
        }
    }
}
