﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net.WebSockets;
using System.Security.Cryptography;
using System.ServiceModel.Syndication;
using System.Text;
using System.Threading.Tasks;
using System.Timers;
using System.Xml;
using Discord;
using Discord.WebSocket;
using Google.Apis.Auth.OAuth2;
using Google.Apis.Sheets.v4;
using Microsoft.EntityFrameworkCore;

namespace DiscoBot.gsmeet
{
    public class GSMeet : IModule
    {
        string IModule.Name { get => "GSMeet"; set => throw new NotImplementedException(); }
        private SocketGuild guild;
        private GSMeetContext gsmeetContext;
        private Dictionary<string, Timer> timers = new Dictionary<string, Timer>();
        private List<WebSocket> webSockets = new List<WebSocket>();

        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        private static readonly string[] Scopes = { SheetsService.Scope.Spreadsheets };
        private static readonly string ApplicationName = "DiscoBot";
        private SheetsService service;

        private List<GSMeeting> lastMeetings = null;



        public GSMeet(SocketGuild guild)
        {
            this.guild = guild;
            gsmeetContext = new GSMeetContext(guild.Id);
            gsmeetContext.Database.EnsureCreated();

            GoogleCredential credential;
            using(var stream = new FileStream("client_secret.json", FileMode.Open, FileAccess.Read))
            {
                credential = GoogleCredential.FromStream(stream).CreateScoped(Scopes);
            }
            service = new SheetsService(new Google.Apis.Services.BaseClientService.Initializer()
            {
                HttpClientInitializer = credential,
                ApplicationName = ApplicationName
            });

            foreach (var f in gsmeetContext.GSheets)
            {
                var gssheet = new GSSheet(f);
                InitializeSheet(gssheet);
                HandleSheetCheck(gssheet);
            }

            Commands.Add("gsmeetadd", HandleGSMeetAddCommand);
            Commands.Add("gsmeetdel", HandleGSMeetDelCommand);
            Commands.Add("gsmeetlist", HandleGSMeetListCommand);
            Commands.Add("gsmeetdebug", HandleGSMeetDebugCommand);
        }

        private Task InitializeSheet(GSSheet sheet)
        {
            Console.WriteLine("Found sheet " + sheet.Db.Name);
            Timer timer = new Timer(sheet.Db.CheckInterval.TotalMilliseconds);
            timer.AutoReset = true;
            timer.Elapsed += async (sender, e) =>
            {
                try
                {
                    await Task.Run(() => HandleSheetCheck(sheet));
                } catch(Exception ex)
                {
                    Console.WriteLine("gsmeet: EXCEPTION IN SHEET CHECK"+ex);
                }
            };
            timer.Start();
            timers.Add(sheet.Db.Name, timer);
            return Task.CompletedTask;
        }

        private Task DeinitializeSheets(GSSheet sheet)
        {
            /*
            Timer t = timers[feed.Name];
            t.Stop();
            timers.Remove(feed.Name);
            */
            return Task.CompletedTask;
        }

        private GSMeeting ParseMeeting(SocketTextChannel c, IList<IList<object>> values)
        {
            var meeting = new GSMeeting();
            var dateRow = values[1];
            var time = TimeSpan.Parse((string)dateRow[0]);
            var idString = "";
            for (int i = 2; i < dateRow.Count; ++i)
            {
                var col = dateRow[i];
                var d = DateTimeOffset.Parse((string)col, null, DateTimeStyles.AssumeUniversal);
                var dateTime = d + time;
                idString += dateTime.ToString("o");
                meeting.Dates.Add(dateTime);
            }
            for(int i=2;i<values.Count;++i)
            {
                var row = values[i];
                var u = new GSMeetingUser();
                u.DiscordTag = (string)row[0];
                try {
                    u.User = guild.Users.Where(gu => (gu.Username + "#" + gu.DiscriminatorValue) == u.DiscordTag).Single();
                } catch(InvalidOperationException e)
                {
                    // Too spammy, do something about it? Don't rely on u.User
                    //c.SendMessageAsync("Could not find User "+ u.DiscordTag+" in Discord, cannot mention.");
                }
                u.Name = (string)row[1];
                for (int j = 2; j < row.Count; ++j)
                {
                    var str = (string)row[j];
                    u.Signups.Add(str);
                }
                // Fill the empty columns that the api skipped
                while(u.Signups.Count < meeting.Dates.Count)
                {
                    u.Signups.Add("");
                }
                meeting.Users.Add(u);
            }
            var sha1 = new SHA1CryptoServiceProvider();
            var binHash = sha1.ComputeHash(Encoding.ASCII.GetBytes(idString));
            var hash = BitConverter.ToString(binHash).Replace("-", string.Empty);
            meeting.Id = hash;
            meeting.FetchTime = DateTimeOffset.Now;
            return meeting;
        }

        private async Task<List<GSMeeting>> ParseGSMeetSheet(GSSheet sheet, IList<IList<object>> values)
        {
            SocketTextChannel c = guild.Channels.Where(g => g.Id == sheet.Db.Channel).Single() as SocketTextChannel;
            int meetStart = -1;
            var val = values as List<IList<object>>;
            var meetings = new List<GSMeeting>();
            for (var i=0;i<values.Count;++i)
            {
                var row = values[i] as List<object>;
                if(row.Count < 1)
                {
                    continue;
                }
                if((string)row[0] == "Raid")
                {
                    if(meetStart == -1)
                    {
                        meetStart = i;
                    } else
                    {
                        meetings.Add(ParseMeeting(c, val.GetRange(meetStart,(i-meetStart-2))));
                        meetStart = i;
                    }
                }
            }
            // Last meeting starts with Raid, thus sets meetStart but no other meeting finishes it, so parse last rows as last meeting
            meetings.Add(ParseMeeting(c, val.GetRange(meetStart, (values.Count - meetStart))));
            Console.WriteLine("I parsed " + meetings.Count + " meetings from the sheet.");
            return meetings;
        }

        private string MissingStringForEvent(GSMeeting m)
        {
            var now = DateTimeOffset.Now;
            string s = "";
            for(var i=0;i<m.Dates.Count;++i)
            {
                var date = m.Dates[i];
                if(date < now || date > now.AddDays(7))
                {
                    continue;
                }
                var us = "";
                foreach (var user in m.Users)
                {
                    if (user.Signups[i] != "1" && user.Signups[i] != "0" && user.Signups[i] != "0.5")
                    {
                        if (user.User != null)
                        {
                            us = user.User.Mention + " ";
                        }
                        else
                        {
                            us += user.Name + " ";
                        }
                    }
                }
                if(us != "")
                {
                    s += date + ": "+ us + "\n";
                }
            }
            return s;
        }

        private async Task HandleSheetCheck(GSSheet sheet)
        {
            SocketTextChannel c = guild.Channels.Where(g => g.Id == sheet.Db.Channel).Single() as SocketTextChannel;
            Console.WriteLine("Checking sheet " + sheet.Db.Id + " :3");
            var values = FetchRangeFromSheet(sheet.Db.Id, $"{sheet.Db.SheetName}!A:I");
            if (values == null || values.Count < 1)
            {
                Console.WriteLine("No values found.");
            }
            lastMeetings = await ParseGSMeetSheet(sheet, values);
            var ln = sheet.Db.LastNotified;
            var now = DateTimeOffset.Now;
            if(ln.AddDays(1) > now)
            {
                Console.WriteLine("It is: "+now+" Next notify is: " + ln.AddDays(1));
                return;
            }
            Console.WriteLine("Notifying...");
            string missStr = "";
            foreach(var meeting in lastMeetings)
            {
                missStr += MissingStringForEvent(meeting);
            }
            if(!string.IsNullOrEmpty(missStr))
            {
                string msg = "Heyo! I found the following missing raid signups:\n";
                msg += missStr;
                msg += "Please sign up so we can raid! :)";
                await c.SendMessageAsync(msg);
            }
            //TODO FIXME: Remove the hardcode 18 and read meeting timespan from sheet and use this value
            var nextDate = now.AddHours(-now.Hour + 18).AddMinutes(-now.Minute).AddSeconds(-now.Second).AddMilliseconds(-now.Millisecond);
            Console.WriteLine("Setting next notify to: " + nextDate);
            sheet.Db.LastNotified = nextDate;
            gsmeetContext.SaveChanges();
            /*
            foreach(var meeting in meetings)
            {
                if(!sheet.Timers.ContainsKey(meeting.Id))
                {
                    sheet.Timers.Add(meeting.Id, new GSMeetingTimers());
                    
                }
                sheet.Timers[meeting.Id].UpdateTimers(c, meeting);
            }
            Console.WriteLine("Parsed sheet " + sheet.Db.Id);
            */
        }

        private Task HandleGSMeetAddCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            string sheetId = parameters[2];
            string sheetName = parameters[3];
            try
            {
                int timeSec = Int32.Parse(parameters[4]);
                DBSheet sheet = new DBSheet();
                sheet.Channel = gchan.Id;
                sheet.Name = name;
                sheet.Id = sheetId;
                sheet.SheetName = sheetName;
                sheet.LastChecked = DateTimeOffset.Now;
                var now = DateTimeOffset.Now;
                sheet.LastNotified = now.AddDays(-1).AddHours(-now.Hour).AddMinutes(-now.Minute).AddSeconds(-now.Second).AddMilliseconds(-now.Millisecond);
                sheet.CheckInterval = new TimeSpan(0, 0, timeSec);
                gsmeetContext.GSheets.Add(sheet);
                try
                {
                    gsmeetContext.SaveChanges();
                    var gssheet = new GSSheet(sheet);
                    InitializeSheet(gssheet);
                    msg.Channel.SendMessageAsync("Sheet " + sheet.Name + " with id " + sheet.Id + " saved.");
                }
                catch (InvalidOperationException)
                {
                    msg.Channel.SendMessageAsync("Unable to save sheet feed.");
                }
            }
            catch (FormatException)
            {
                msg.Channel.SendMessageAsync("Unable to save sheet. Invalid check time.");
            }
            return Task.CompletedTask;
        }

        private Task HandleGSMeetListCommand(SocketMessage msg, string[] parameters)
        {
            /*
            List<RssFeed> feeds = rssContext.RssFeeds.Where(f => f.Channel == msg.Channel.Id).ToList();
            List<string> m = new List<string>();
            m.Add("Feeds for this channel are:");
            m.Add("Name | URL | CheckInterval | LastChecked");
            msg.Channel.SendMessageAsync(string.Join("\n", m));
            m = new List<string>();
            int cnt = 0;
            foreach (var f in feeds)
            {
                m.Add("- " + f.Name + " | " + f.Url + " | " + f.CheckInterval + " | " + f.LastChecked);
                cnt++;
                if(cnt > 4)
                {
                    cnt = 0;
                    msg.Channel.SendMessageAsync(string.Join("\n", m));
                }
            }
            if(cnt > 0)
                msg.Channel.SendMessageAsync(string.Join("\n", m));
                */
            return Task.CompletedTask;
        }

        private Task HandleGSMeetDelCommand(SocketMessage msg, string[] parameters)
        {
            /*
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            try
            {
                RssFeed f = rssContext.RssFeeds.Where(f => f.Name == name).Single();
                rssContext.Remove(f);
                rssContext.SaveChanges();
                DeinitializeFeed(f);
                msg.Channel.SendMessageAsync("Removed feed " + f.Name);
            } catch(InvalidOperationException)
            {
                msg.Channel.SendMessageAsync("Could not find feed " + name);
            }
            */
            return Task.CompletedTask;
        }

        private Task HandleGSMeetDebugCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            msg.Channel.SendMessageAsync("Username: "+msg.Author.Username+"#"+msg.Author.DiscriminatorValue);
            return Task.CompletedTask;
        }

        private IList<IList<object>> FetchRangeFromSheet(string sheetId, string range)
        {
            var request = service.Spreadsheets.Values.Get(sheetId, range);
            var response = request.Execute();
            return response.Values;
        }

        public void Initialize()
        {
            Console.WriteLine("Initializing gsmeet...");
        }

        public async Task OnNewWebSocketAsync(WebSocket ws, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Calendar " + guild.Id + " has a new websocket.");
            webSockets.Add(ws);
        }
    }
}
