﻿using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Timers;
using System.Linq;
using System.Threading.Tasks;
using System.ComponentModel.DataAnnotations.Schema;
using Discord.WebSocket;

namespace DiscoBot.gsmeet
{
    public class GSMeetContext : DbContext
    {
        private ulong guildId;
        public DbSet<DBSheet> GSheets { get; set; }

        public GSMeetContext(ulong guildId)
        {
            this.guildId = guildId;
        }

        protected override void OnConfiguring(DbContextOptionsBuilder optionsBuilder)
        {
            optionsBuilder.UseSqlite("Data Source=" + guildId + "-gsmeet.db");
        }
    }

    public class DBSheet
    {
        [Key]
        public string Name { get; set; }
        public string Id { get; set; }
        public string SheetName { get; set; }
        public ulong Channel { get; set; }
        public TimeSpan CheckInterval { get; set; }
        public DateTimeOffset LastChecked { get; set; }
        public DateTimeOffset LastNotified { get; set; }
    }

    public class GSSheet
    {
        public GSSheet(DBSheet db)
        {
            this.Db = db;
        }
        public DBSheet Db { get; set; }
    }

    public class GSMeetingUser
    {
        public GSMeetingUser()
        {
            this.Signups = new List<string>();
        }
        public string DiscordTag { get; set; }
        public SocketGuildUser User { get; set; }
        public string Name { get; set; }
        public List<string> Signups { get; set; }
    }

    public class GSMeeting
    {
        public GSMeeting()
        {
            this.Users = new List<GSMeetingUser>();
            this.Dates = new List<DateTimeOffset>();
        }
        public string Id { get; set; }
        public DateTimeOffset FetchTime { get; set; }
        public List<GSMeetingUser> Users { get; set; }
        public List<DateTimeOffset> Dates { get; set; }
    }
}
